#
# Copyright (C) 2006 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

__all__ = ['AnalyzeThread',
           'PluginReportReceiver',
           'SETroubleshootDatabase',
           'LogfileAnalyzer',
          ]

PROGNAME="setroubleshoot"
import gettext
gettext.bindtextdomain(PROGNAME, "/usr/share/locale")
gettext.textdomain(PROGNAME)
try:
    gettext.install(PROGNAME, "/usr/share/locale", 1)
except IOError:
    import __builtin__
    __builtin__.__dict__['_'] = unicode

from setroubleshoot.avc_audit import *
from setroubleshoot.config import cfg
from setroubleshoot.errcode import *
from setroubleshoot.log import *
from setroubleshoot.rpc import *
from setroubleshoot.rpc_interfaces import *
from setroubleshoot.signature import *
from setroubleshoot.util import *

import errno
import gobject
import os
import Queue
import signal
from stat import *
import re
import sys
import time
import threading
from types import *

#------------------------------------------------------------------------------

class AVCInfo(object):
    def __init__(self, avc, query_environment=True, line_numbers=None, timestamp=None):
        self.avc = avc
        self.query_environment = query_environment
        self.line_numbers = line_numbers
        if timestamp is None:
            self.timestamp = TimeStamp()
        else:
            self.timestamp = timestamp
        self.template_substitutions = {}
        self.path = None
        self.binary = None
        self.source_pkg = None
        self.src_rpms=[]
        self.tgt_rpms=[]

    def update(self):
        self.derive_axillary_info()
        self.update_derived_template_substitutions()

    def derive_axillary_info(self):
        # Do basic processing of the avc to extract information that
        # most plugins will use. Only information that requires
        # processing or is part of the signature should be handled here
        # (e.g., determing the exe name). Simple information that can
        # unambiogously be retrieved from the original avc is not usually
        # processed (e.g., source type).
        #
        # The current fields that are processed and stored are.
        # * binary (either exe or comm)
        # * path (name or path)
        # * package

        self.path = self.avc.get_path()
        self.binary = self.avc.get_binary()
        self.src_rpms=[]
        self.tgt_rpms=[]

        if self.query_environment:
            if self.binary is not None:
                self.source_pkg = get_rpm_nvr_by_file_path(self.binary)
                if self.source_pkg is not None:
                    self.src_rpms.append(self.source_pkg)
        
            if self.path is not None:
                rpm = get_rpm_nvr_by_file_path(self.path)
                if rpm is not None:
                    self.tgt_rpms.append(rpm)


    def set_alt_path(self, path):
        if self.path is None:
            self.path = path

    def set_template_substitutions(self, **kwds):
        for key, value in kwds.items():
            if value:
                self.template_substitutions[key] = value

    def set_boolean(self, bool_name):
        self.set_template_substitutions(BOOLEAN=bool_name)

    def update_derived_template_substitutions(self):
        self.template_substitutions["SOURCE_TYPE"] = self.avc.scontext.type
        self.template_substitutions["TARGET_TYPE"] = self.avc.tcontext.type
        self.template_substitutions["SOURCE_PATH"] = self.binary
        self.template_substitutions["TARGET_PATH"] = self.path

        if self.path is None:
            self.template_substitutions["TARGET_DIR"] = None
        else:
            self.template_substitutions["TARGET_DIR"] = os.path.dirname(self.path)

        self.template_substitutions["TARGET_CLASS"] = self.avc.tclass

        if self.avc.access is None:
            self.template_substitutions["ACCESS"] = None
        else:
            self.template_substitutions["ACCESS"] = ' '.join(self.avc.access)

        self.template_substitutions["SOURCE_PACKAGE"] = self.source_pkg
        if self.avc.dest is None:
            self.template_substitutions["PORT_NUMBER"] = self.avc.src
        else:
            self.template_substitutions["PORT_NUMBER"] = self.avc.dest

    def validate_template_substitutions(self):
        # validate, replace any None values with friendly string
        for key, value in self.template_substitutions.items():
            if value is None:
                self.template_substitutions[key] = _('<Unknown>')

#------------------------------------------------------------------------------

class PluginStatistics(object):
    def __init__(self, plugin):
        self.name = plugin.analysis_id
        self.analyze_start_time = None
        self.analyze_end_time = None
        self.analyze_elapsed_time = None
        self.report_start_time = None
        self.report_end_time = None
        self.report_elapsed_time = None

    def __str__(self):
        analyze_elapsed_time = format_elapsed_time(self.analyze_elapsed_time)

        if self.report_elapsed_time is None:
            return "%s: %s elapsed" % (self.name, analyze_elapsed_time)
        else:
            total_elapsed_time = format_elapsed_time(self.report_end_time-self.analyze_start_time)
            report_elapsed_time = format_elapsed_time(self.report_elapsed_time)

            return "%s: %s elapsed, %s analyze elapsed, %s report elapsed" % \
                   (self.name, total_elapsed_time, analyze_elapsed_time, report_elapsed_time)

    def analyze_start(self):
        self.analyze_start_time = time.time()

    def analyze_end(self):
        self.analyze_end_time = time.time()
        self.analyze_elapsed_time = self.analyze_end_time - self.analyze_start_time

    def report_start(self):
        self.report_start_time = time.time()

    def report_end(self):
        self.report_end_time = time.time()
        self.report_elapsed_time = self.report_end_time - self.report_start_time

#------------------------------------------------------------------------------

class AnalyzeStatistics(object):
    def __init__(self, num_plugins):
        self.num_plugins = num_plugins
        self.cur_plugin = None
        self.called_plugins = []
        self.start_time = None
        self.end_time = None
        self.elapsed_time = None

    def __str__(self):
        elapsed_time = None
        avg_plugin_time = None
        n_called = len(self.called_plugins)
        if self.elapsed_time is not None:
            elapsed_time = format_elapsed_time(self.elapsed_time)
            if n_called:
                avg_plugin_time = format_elapsed_time(self.elapsed_time / n_called)

        return "%d/%d plugins in %s elapsed, avg plugin %s elapsed, plugins=[\n%s\n]" % \
               (n_called, self.num_plugins, elapsed_time, avg_plugin_time,
                self.called_plugins_to_string())

    def called_plugins_to_string(self):
        return '\n'.join([str(x) for x in self.called_plugins])
                                               
    def start(self):
        self.start_time = time.time()

    def end(self):
        self.end_time = time.time()
        self.elapsed_time = self.end_time - self.start_time

    def new_plugin(self, plugin):
        self.cur_plugin = PluginStatistics(plugin)
        self.called_plugins.append(self.cur_plugin)
        self.cur_plugin.analyze_start()

#------------------------------------------------------------------------------

class Analyze(object):
    def __init__(self):
        self.plugins = load_plugins()
        if debug:
            log_avc.debug("Number of Plugins = %d", len(self.plugins))
        
    # FIXME: perhaps the avc parameter should be an AuditEvent instead

    def analyze_avc(self, avc, report_receiver, query_environment=True, line_numbers=None, timestamp=None):
        if debug:
            log_avc.debug("analyze_avc() avc=%s", avc)

        avcinfo = AVCInfo(avc, query_environment, line_numbers, timestamp)
        avcinfo.update()

        if profile:
            statistics = AnalyzeStatistics(len(self.plugins))
            statistics.start()

        for plugin in self.plugins:
            if profile:
                statistics.new_plugin(plugin)
                statistics.cur_plugin.analyze_start()
            try:	
                report = plugin.analyze(avcinfo)
                if report is not None:
                    if profile:
                        statistics.cur_plugin.analyze_end()
                    report_receiver.report_problem(report)
                    break;

            except Exception, e:
                log_avc.exception("Plugin Exception %s " % plugin.analysis_id) 
                self.plugins.remove(plugin)

            if profile:
                statistics.cur_plugin.analyze_end()

        if profile:
            statistics.end()
            log_stats.info("analyze_avc() avc=%s\nstatistics=%s", avc, statistics)
    
#------------------------------------------------------------------------------

class AnalyzeThread(Analyze, threading.Thread):
    def __init__(self,queue):
        # parent class constructors
        threading.Thread.__init__(self)
        Analyze.__init__(self)

        self.queue=queue
        
    def run(self):
        while True:
            event, report_receiver = self.queue.get()
            event.avc=AVC(event.audit_dict)
            self.analyze_avc(event.avc, report_receiver,
                             timestamp=TimeStamp(event.timestamp))

#------------------------------------------------------------------------------

class PluginReportReceiver(object):
    def __init__(self, database):
        self.database = database

    def report_problem(self, siginfo):
        try:
            database_siginfo = self.database.lookup_signature(siginfo.sig)
            database_siginfo.update_merge(siginfo)
            self.database.modify_siginfo(database_siginfo)
            if debug:
                log_database.debug("signature found in database")
        except ProgramError, e:
            if e.errno == ERR_NO_SIGNATURE_MATCH:
                if debug:
                    log_database.debug("not in database yet")
                siginfo.local_id=self.database.sigs.generate_local_id()
                siginfo.first_seen_date = siginfo.last_seen_date
                database_siginfo = self.database.add_siginfo(siginfo)
            else:
                raise

        database_siginfo.report_count += 1
        return database_siginfo


#------------------------------------------------------------------------------

class SETroubleshootDatabase(object):
    def __init__(self, name, friendly_name=None, notify=None, temp=False, replace=False):
        self.filepath = make_database_filepath(name, temp)
        self.notify = notify
        self.properties = SEDatabaseProperties(name=name, friendly_name=friendly_name, filepath=self.filepath)
        self.lock = threading.Lock()
        self.modified_count = 0
        self.auto_save_interval = 30   # auto save after this many seconds
        self.auto_save_threshold = 200 # auto save after this many changes
        self.auto_save_timer = None

        if debug:
            log_database.debug("created new database: name=%s, friendly_name=%s, filepath=%s",
                               self.properties.name, self.properties.friendly_name, self.properties.filepath)
        if replace:
            self.remove()

        self.load()

    def load(self):
        self.sigs = SEFaultSignatureSet()
        
        if os.path.exists(self.filepath):
           stat_info = os.stat(self.filepath)
           if stat_info[ST_SIZE] > 0:
               self.sigs.read_xml_file(self.filepath, 'sigs')

    def save(self):
        if debug:
            log_database.debug("writing database (%s) modified_count=%s",
                               self.filepath, self.modified_count)
        self.sigs.write_xml('sigs', self.filepath)
        self.modified_count = 0
        if self.auto_save_timer is not None:
            gobject.source_remove(self.auto_save_timer)
            self.auto_save_timer = None
        
    def mark_modified(self):
        self.modified_count += 1
        if self.modified_count > self.auto_save_threshold:
            self.save()
        elif self.auto_save_timer is None:
            self.auto_save_timer = \
                gobject.timeout_add(self.auto_save_interval*1000,
                                    self.auto_save_callback)

    def auto_save_callback(self):
        if debug:
            log_database.debug("auto_save database (%s) modified_count=%s",
                               self.filepath, self.modified_count)
        self.save()
        return False

    def remove(self):
        if os.path.exists(self.filepath):
            if debug:
                log_database.debug("deleting database (%s)", self.filepath)
            os.remove(self.filepath)

    def acquire(self):
        self.lock.acquire()

    def release(self):
        self.lock.release()
    
    def lookup_signature(self, sig):
        siginfo = None

        matches = self.sigs.match_signatures(sig)
        if debug:
            log_sig.debug("lookup_signature: found %d matches with scores %s",
                          len(matches), ",".join(["%.2f" % x.score for x in matches]))
        if len(matches) == 0:
            raise ProgramError(ERR_NO_SIGNATURE_MATCH)
        elif len(matches) > 1:
            raise ProgramError(ERR_MULTIPLE_SIGNATURE_MATCH)
        siginfo = matches[0].siginfo
        return siginfo

    def lookup_local_id(self, local_id):
        siginfo = self.sigs.lookup_local_id(local_id)
        if debug:
            log_database.debug("lookup_local_id: %s found %s", local_id, siginfo)
        if siginfo is None:
            if debug:
                log_database.debug("lookup_local_id: %s not found", local_id)
            raise ProgramError(ERR_SIGNATURE_ID_NOT_FOUND, "id (%s) not found" % local_id)
        return siginfo

    def add_siginfo(self, siginfo):
        siginfo = self.sigs.add_siginfo(siginfo)
        if self.notify:
            self.notify.signatures_updated('add', siginfo.local_id)
        self.mark_modified()
        return siginfo

    def query_alerts(self, criteria):
        if debug:
            log_database.debug("query_alerts: criteria=%s", criteria)

        if criteria == '*':
            return self.sigs

        # FIXME: we assume if criteria is not wildcard its a local_id, need a more general/robust mechanism
        sigs = SEFaultSignatureSet()
        siginfo = self.lookup_local_id(criteria)
        sigs.add_siginfo(siginfo)
        return sigs

    def delete_signature(self, sig):
        if debug:
            log_database.debug("delete_signature: sig=%s", sig)

        siginfo = self.lookup_signature(sig)
        self.sigs.remove_siginfo(siginfo)
        if self.notify:
            self.notify.signatures_updated('delete', siginfo.local_id)
        self.mark_modified()
        
    def modify_siginfo(self, siginfo):
        if self.notify:
            self.notify.signatures_updated('modify', siginfo.local_id)
        self.mark_modified()


    def evaluate_alert_filter(self, sig, username):
        if debug:
            log_database.debug("evaluate_alert_filter: username=%s sig=%s", username, sig)

        siginfo = self.lookup_signature(sig)
        action = siginfo.evaluate_filter_for_user(username)
        return action

    def set_user_data(self, sig, username, item, data):
        if debug:
            log_database.debug("set_user_data: username=%s item=%s data=%s sig=\n%s",
                               username, item, data, sig)

        siginfo = self.lookup_signature(sig)
        user_data = siginfo.get_user_data(username)
        user_data.update_item(item, data)
        self.mark_modified()

    def set_filter(self, sig, username, filter_type, data = "" ):
        if debug:
            log_database.debug("set_filter: username=%s filter_type=%s sig=\n%s",
                               username, filter_type, sig)

        siginfo = self.lookup_signature(sig)
        siginfo.update_user_filter(username, filter_type, data)
        self.mark_modified()

    def add_user(self, username):
        self.user = self.sigs.users.add_user(username)
        self.mark_modified()
        
    def get_user(self, username):
        return self.sigs.users.get_user(username)
        
#------------------------------------------------------------------------------

class SETroubleshootDatabaseLocal(RpcManage,
                                  SETroubleshootDatabaseInterface,
                                  gobject.GObject):

    __gsignals__ = {
        'signatures_updated': 
        (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT, gobject.TYPE_PYOBJECT)),
        }


    def __init__(self, database, notify=None):
        gobject.GObject.__init__(self)
        RpcManage.__init__(self)
        self.database = database
        if notify is not None:
            self.database.notify = notify

    def emit_rpc(self, rpc_id, type, rpc_def, *args):
        if debug:
            log_rpc.debug("%s emit %s(%s) id=%s", self.__class__.__name__,
                          rpc_def.method, ','.join([str(arg) for arg in args]), rpc_id)
        async_rpc = self.async_rpc_cache[rpc_id]
        func = getattr(self.database, rpc_def.method, None)
        print "func=%s" % (repr(func))
        if func is None:
            raise ProgramError(ERR_METHOD_NOT_FOUND,
                               "method %s not found in base class of %s" % (rpc_def.method, self.__class__.__name__))
        try:
            async_rpc.return_args = func(*args)
            async_rpc.return_type = 'method_return'
            if async_rpc.return_args is not None:
                async_rpc.return_args = [async_rpc.return_args]
        except ProgramError, e:
            async_rpc.return_args = [e.errno, e.strerror]
            async_rpc.return_type = 'error_return'

        if async_rpc.return_args is not None:
            gobject.idle_add(self.process_async_return, async_rpc)


gobject.type_register(SETroubleshootDatabaseLocal)

#------------------------------------------------------------------------------

class LogfileAnalyzer(object):
    def __init__(self, logfile_path, notify=None):
        self.logfile_path = logfile_path

        self.logfile_basename = os.path.basename(logfile_path)
        self.friendly_name = "file: %s" % (os.path.splitext(self.logfile_basename)[0])
        self.database = SETroubleshootDatabase(self.logfile_basename, notify=notify,
                                               friendly_name=self.friendly_name, temp=True, replace=True)
        self.database_rpc = SETroubleshootDatabaseLocal(self.database)

        self.report_receiver = PluginReportReceiver(self.database)

    
    def analyze_logfile(self, progress_callback=None):
        if debug:
            log_avc.debug('analyze_logfile: file=%s', self.logfile_path)

        analyze = Analyze()

        def avc_event_handler(event):
            if debug:
                log_avc.debug('avc_event_handler() event=%s', event)
            if event.is_avc and not event.is_granted and len(event.audit_dict) > 0:
                event.avc=AVC(event.audit_dict)
                analyze.analyze_avc(event.avc, self.report_receiver, False,
                                    line_numbers=event.line_numbers.keys(),
                                    timestamp=TimeStamp(event.timestamp))


        statistics = parse_audit_logfile(self.logfile_path, avc_event_handler,
                                         self.report_receiver, progress_callback)
        #self.database.save() # FIXME

        if profile:
            log_stats.info('analyze_logfile: stats=%s', statistics)
        if debug:
            #log_avc.debug('analyze_logfile: results=%s', logfile_database.query_alerts('*'))
            pass

        
