#
# Copyright (C) 2006 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

import audit
import AuditMsg
import os
import re
import select
import selinux
import socket as Socket
from stat import *
import struct
import sys
import thread
import threading
import time

from setroubleshoot.config import cfg
from setroubleshoot.errcode import *
from setroubleshoot.log import *
from setroubleshoot.signature import AVC, AvcContext
from setroubleshoot.util import *

__all__ = [
    'AuditSocketReceiverThread',
    'parse_audit_logfile',
    ]

#------------------------------ Variables -------------------------------

audit_type_map_to_int = {
    'AVC'      : audit.AUDIT_AVC,
    'AVC_PATH' : audit.AUDIT_AVC_PATH,
    'SYSCALL'  : audit.AUDIT_SYSCALL,
    'CWD'      : audit.AUDIT_CWD,
    'PATH'     : audit.AUDIT_PATH,
}

audit_type_map_to_str = {
    audit.AUDIT_AVC      : 'AVC',
    audit.AUDIT_AVC_PATH : 'AVC_PATH',
    audit.AUDIT_SYSCALL  : 'SYSCALL',
    audit.AUDIT_CWD      : 'CWD',
    audit.AUDIT_PATH     : 'PATH',
}

#------------------------------------------------------------------------------

my_context = AvcContext(selinux.getcon()[1])
def verify_avc(avc):
    if avc.scontext.type == None or avc.tcontext.type == None:
        return False

    if my_context.type == avc.scontext.type:
        log_program.error( "Can not handle AVC'S related to dispatcher. exiting\nsetroubleshoot context=%s, AVC scontext=%s", my_context, avc.scontext)
        sys.exit(0)

    return True

#------------------------------------------------------------------------------

class LogfileStatistics(object):
    def __init__(self, logfile_path):
        self.logfile_path = logfile_path
        self.msg_count = 0
        self.line_count = 0
        self.event_count = 0
        self.max_cache_length = 0
        self.start_time = None
        self.end_time = None
        self.elapsed_time = None

    def __str__(self):
        elapsed_time = None
        avg_event_time = None
        if self.elapsed_time is not None:
            elapsed_time = format_elapsed_time(self.elapsed_time)
            if self.event_count:
                avg_event_time = format_elapsed_time(self.elapsed_time / self.event_count)
        return "%s: %d lines, %d messages, %d events, max cache length %d, %s elapsed, avg event %s elapsed" % \
          (self.logfile_path, self.line_count, self.msg_count, self.event_count,
           self.max_cache_length, elapsed_time, avg_event_time)

    def start(self):
        self.start_time = time.time()

    def end(self):
        self.end_time = time.time()
        self.elapsed_time = self.end_time - self.start_time

#------------------------------------------------------------------------------
'''
audit messages are prefixed with an event id which looks like this:

audit(1152828331.857:123085)

All audit messages belonging to one event share the exact same event id
and that id is guaranteed to be unique to that event.

There are 3 fields:

The first two fields which are separated by a '.' consititue a
timestamp. The first integer is the number of seconds since the
beginning of the UNIX epoch and the second integer is the number of
milliseconds. Together they form a high precision timestamp. The 3rd
integer is used to make the event unique within the timestamp.

FYI: What happens if the clock ticks between audit messages?

From the PoV of an AVC, what really happens is this...a thread makes a call 
into the kernel. The audit system resets a special memory area for that 
thread and stuffs some values into the data structure in case they are needed 
later. The syscall goes about its business. Eventually it does an avc 
calculation that fails. The selinux kernel code calls the audit system to 
report the event. As it processes the message, the audit system looks to see 
if there is a time stamp. If not it gets one and marks the syscall as 
auditable. The avc goes out immediately using that time stamp. The syscall 
continues on its way until it reaches syscall exit. The audit system sees 
that the syscall was marked auditable and proceeds to dump everything that it 
collected during that syscall excursion using the stored time stamp.
'''
#------------------------------------------------------------------------------

def parse_audit_logfile(filepath, avc_event_handler, report_receiver, progress_callback=None):
    # regular expression to match lines like this:
    # type=AVC msg=audit(1152828325.857:123085): avc:  denied ...
    # group 1 is the type, everything after that is the message body
    audit_msg_type_re = re.compile('^type=([^\s]+)\s+')

    audit_receiver = AuditMsgReceiver(avc_event_handler)
    msg_count = 0
    flush_count = 0
    flush_size = 30
    start_time = time.time()

    try:
        f = open(filepath)
    except IOError, e:
        raise ProgramError(ERR_FILE_OPEN, detail="%s, %s" % (filepath, e.strerror))
    if profile:
        statistics = LogfileStatistics(filepath)
    else:
        statistics = None
    stat_info = os.stat(filepath)
    file_size = stat_info[ST_SIZE]
    n_bytes_read = 0
    progress = 0.0
    progress_increment = 0.01
    next_progress_interval = progress + progress_increment
    line_number = 0

    if profile:
        statistics.start()
    for line in f:
        line_number += 1
        n_bytes_read += len(line)
        progress = float(n_bytes_read) / float(file_size)
        if progress >= next_progress_interval or n_bytes_read == file_size:
            if progress_callback:
                progress_callback(progress)
            next_progress_interval = progress + progress_increment
                
        
        match = audit_msg_type_re.search(line)
        if match:
            # Found an audit message
            msg_type_str = match.group(1)

            # Map the type string to one of the type constants
            msg_type_num = audit_type_map_to_int.get(msg_type_str, None)

            # If its a type we recognize, then proceed
            if msg_type_num is not None:
                msg_count += 1
                flush_count += 1

                # Inject the message into the receiving system
                audit_receiver.feed(msg_type_num, line[match.end():], line_number)

                # If we've seen enough messages then sweep the event
                # cache and flush what we can
                if flush_count > flush_size:
                    audit_receiver.flush()
                    flush_count = 0

    # All done, flush all buffered events out
    audit_receiver.close()

    if profile:
        statistics.end()
        statistics.line_count = line_number
        statistics.msg_count = msg_count
        statistics.event_count = audit_receiver.event_count
        statistics.max_cache_length = audit_receiver.max_cache_length
        audit_receiver.reset_statistics()

    if n_bytes_read != file_size:
        log_avc.warn("failed to read complete file, %d bytes read out of total %d bytes (%s)",
                     n_bytes_read, file_size)
        if progress_callback:
            progress_callback(1.0)

    return statistics
                
#------------------------------------------------------------------------------

class AuditEvent:
    # regular expression to find an avc message, e.g.:
    # avc:  denied  { append } for  pid=14205  ...
    # group 1 is the grant, e.g. denied
    # group 2 is the access list, e.g. append
    avc_re = re.compile("avc:\s+([^\s]+)\s+{([^}]+)}\s+for\s+")

    # regular expression to find key value pairs of the form "key = value"
    # group 1 is the key
    # group 2 is the value
    key_value_pair_re = re.compile("([^ \t]+)\s*=\s*([^ \t]+)")

    def __init__(self, msg_id, timestamp, msg):
        self.msg_id       = msg_id
        self.timestamp    = timestamp
        self.msg          = msg
        self.audit_dict   = {}
        self.is_avc       = False
        self.is_granted   = False
        self.line_numbers = {}
        self.avc          = None

    def __str__(self):
        keys = self.audit_dict.keys()
        keys.sort()
        return "msg_id=%s, timestamp=%f, is_avc=%s, is_granted=%s: line_numbers=[%s] %s" % \
               (self.msg_id, self.timestamp, self.is_avc, self.is_granted,
                ",".join([str(x) for x in self.line_numbers.keys()]),
                ",".join(["%s=%s" % (key, self.audit_dict[key]) for key in keys]))

    def add_msg(self, msg_type, msg):
        '''add the contents of an audit message to this event'''
        if msg_type == audit.AUDIT_AVC:
            match = self.avc_re.search(msg)
            if match is None:
                raise ValueError("cannot not parse AVC message: '%s'" % msg)
            grant = match.group(1)
            if grant == 'denied':
                is_granted = False
            else:
                is_granted = True
            access_str = match.group(2)
            access_list = access_str.split()
            self.audit_dict['access'] = access_list
            msg = msg[match.end():]
            self.is_avc = True

        for match in self.key_value_pair_re.finditer(msg):
            key   = match.group(1)
            value = match.group(2)
            self.audit_dict[key] = value

#------------------------------------------------------------------------------

class AuditMsgReceiver:
    """
    The audit system emits messages about a single event
    independently. Thus one single auditable event may be composed
    from one or more individual audit messages. Each audit message is
    prefixed with a unique event id, which includes a timestamp. The
    last audit message associated with an event is not marked in any
    fashion. Audit messages for a specific event may arrive
    interleaved with audit messages for other events. It is the job of
    higher level software (this code) to assemble the audit messages
    into events. The AuditEvent class is used for assembly. When a new
    event id is seen a new AuditEvent object is created, then
    every time an audit message arrives with that event id it is added
    to that object. The AuditEvent object contains the timestamp
    associated with the audit event as well as other data items useful
    for processing and handling the event.

    The audit system does not tell us when the last message belonging
    to an event has been emitted so we have no explicit way of knowing
    when the audit event has been fully assembled from its constituent
    message parts. We use the heuristic if a sufficient length of
    time has expired since we last saw a message for this event, then
    it must be complete

    Thus when audit events are created we place them in a cache where
    they will reside until their time to live has expired at which
    point we will assume they are complete and emit the event.

    Events are expired in the flush_cache() method. The events
    resident in the cache are sorted by their timestamps. A time
    threshold is established. Any events in the cache older than the
    time threshold are flushed from the cache as complete events.

    When should flushes be performed? The moment when a new message is
    added would seem a likely candidate moment to perform a sweep of
    the cache. But this is costly and does not improve how quickly
    events are expired. We could wait some interval of time (something
    much greater than how long we expect it would take for messages
    percolate) and this has good behavior, except for the following
    case. Sometimes messages are emitted by audit in rapid
    succession. If we swept the cache once a second then the cache may
    have grown quite large. Since it is very likely that any given audit
    event is complete by the time the next several events start
    arriving we can optimize by tracking how many messages have
    arrived since the last time we swept the cache.

    The the heuristic for when to sweep the cache becomes:

    If we've seen a sufficient number of messages then sweep -or- if
    a sufficient length of time has elapsed then we sweep

    Note that when audit messages are injected via log file scanning
    elapsed wall clock time has no meaning relative to when to perform
    the cache sweep. However, the timestamp for an event remains a
    critical factor when deciding if an event is complete (have we
    scanned far enough ahead such we're confident we won't see any
    more messages for this event?). Thus the threshold for when to
    expire an event from the cache during static log file scanning is
    determined not by wall clock time but rather by the oldest
    timestamp in the cache (e.g.there is enough spread between
    timestamps in the cache its reasonable to assume the event is
    complete). One might ask in the case of log file scanning why not
    fill the cache until EOF is reached and then sweep the cache?
    Because in log files it is not unusual to have thousands or tens
    of thousands of events and the cache would grown needlessly
    large. Because we have to deal with the real time case we already
    have code to keep only the most recent events in the cache so we
    might as well use that logic, keep the code paths the same and
    minimize resource usage.
    """

    # number of seconds an event must reside in the cache until its
    # considered complete
    cache_time_to_live = 0.005

    # regular expression to find message like this:
    # msg=audit(1152828325.857:123085): avc:  denied  { append } for  pid=14205 ...
    # Note, messages arriving directly from the audit system omit
    # 'msg=', but messages in log files prepend 'msg='
    # group 1 is the optional 'msg='
    # group 2 is the complete event id
    # group 3 is the seconds component of the timestamp
    # group 4 is the millisconds component of the timestamp
    # group 5 is the timestamp unique number
    # group 6 is the body of the message appearing after the event id
    audit_msg_re = re.compile('^\s*(msg=)?audit\(((\d+)\.(\d+):(\d+))\):\s(.*)')

    def __init__(self, event_callback):
        self.event_callback = event_callback
        self.cache = {}
        self.reset_statistics()

    def reset_statistics(self):
        self.max_cache_length = 0
        self.event_count = 0

    def add_msg_to_cache(self, msg_type, msg_id, timestamp, msg, line_number=None):
        if debug:
            log_avc.debug("add_msg_to_cache(): msg_type=%s msg_id=%s line_number=%s",
                          msg_type, msg_id, line_number)
        event = self.cache.setdefault(msg_id, AuditEvent(msg_id, timestamp, msg))
        if line_number is not None:
            event.line_numbers[line_number] = 1
        event.add_msg(msg_type, msg)
        
    def emit_event(self, event):
        self.event_count += 1
        self.event_callback(event)

    def flush_cache(self, threshold_age=None):
        '''Flush events from the cache if they are older than the threshold age.
        If the threshold age is None then the threshold age is set to the age
        of the newest event in the cache minus the cache time to live, in other
        words anything in the cache which is older than the time to live relative
        to the most current event is flushed.
        '''

        # no events, nothing to do
        if len(self.cache) == 0:
            return

        if len(self.cache) > self.max_cache_length:
            self.max_cache_length = len(self.cache)
            
        # flush everything
        if threshold_age == 0:
            for event in self.cache.values():
                self.emit_event(event)
            self.cache = {}
            return

        # flush old events
        ids = self.cache.keys()
        ids.sort(lambda a,b: cmp(self.cache[a].timestamp, self.cache[b].timestamp))
        if threshold_age is None:
            threshold_age = self.cache[ids[-1]].timestamp - self.cache_time_to_live

        for i in ids:
            event = self.cache[i]
            if event.timestamp < threshold_age:
                self.emit_event(event)
                del(self.cache[i])
                

    def flush(self, threshold_age=None):
        self.flush_cache(threshold_age)

    def close(self):
        """Emit every event in the cache irrespective of its
        timestamp. This means we're done, nothing should remain buffered."""
        
        self.flush(0)

    def feed(self, msg_type, msg, line_number=None):
        """Accept a new audit message into the system for processing.
        The caller must supply the message type and the body of the message"""
        if debug:
            log_avc.debug("feed() got %s: '%s'",
                          audit_type_map_to_str.get(msg_type, msg_type), msg)
        match = self.audit_msg_re.search(msg)
        if match is None:
            raise ValueError("cannot parse audit message: '%s'" % msg)
        msg_id       = match.group(2)
        seconds      = int(match.group(3))
        milliseconds = int(match.group(4))
        time_id      = int(match.group(5))
        msg          = match.group(6)

        if msg_type not in (audit.AUDIT_AVC, audit.AUDIT_AVC_PATH,
                            audit.AUDIT_SYSCALL, audit.AUDIT_CWD,
                            audit.AUDIT_PATH):
            return

        timestamp = float(seconds) + milliseconds / 1000.0
        self.add_msg_to_cache(msg_type, msg_id, timestamp, msg, line_number)

#------------------------------------------------------------------------------

class AuditSocketReceiverThread(threading.Thread):
    def __init__(self, queue, report_receiver):
        # parent class constructor
        threading.Thread.__init__(self)
        self.queue = queue
        self.report_receiver = report_receiver
        self.msg_receiver = AuditMsgReceiver(self.handle_avc_event)
        self.audit_socket_path = cfg.get('audit','socket_path')
        self.retry_interval = cfg.getint('audit','retry_interval')

    def connect(self):
        while True:
            try:
                self.audit_socket=Socket.socket(Socket.AF_UNIX,Socket.SOCK_STREAM)
                self.audit_socket.connect(self.audit_socket_path)
                log_avc.info("audit socket (%s) connected", self.audit_socket_path)
                return
            except Socket.error, e:
                log_avc.warning("audit socket (%s) failed, error='%s', retry in %d seconds",
                                self.audit_socket_path, e[1], self.retry_interval)
                
            except OSError, e:
                log_avc.warning("audit socket (%s) failed, error='%s', retry in %d seconds",
                                self.audit_socket_path, e[1], self.retry_interval)

            time.sleep(self.retry_interval)

    def handle_avc_event(self, event):
        if debug:
            log_avc.debug("handle_avc_event: event=%s", event)
        if event.is_avc and not event.is_granted and len(event.audit_dict) > 0:
            event.avc=AVC(event.audit_dict)
            if verify_avc(event.avc):
                self.queue.put((event, self.report_receiver))


    def run(self):
        msg_count = 0
        flush_count = 0
        flush_size = 100
        
        self.connect()
        
        while True:
            inList, outList, errList = select.select([self.audit_socket],[], [], 2)
            try:
                if self.audit_socket in inList:
                    msg = AuditMsg.AuditMsg()
                    if not msg.read_from_fd(self.audit_socket.fileno()):
                        if debug:
                            log_avc.debug("audit socket connection dropped")
                        self.connect()
                    else:
                        msg_count += 1
                        flush_count += 1
                        self.msg_receiver.feed(msg.get_msg_type(), msg.get_body())
                        if flush_count > flush_size:
                            self.msg_receiver.flush()
                            flush_count = 0
                else:
                    # timeout, anything waiting in our event cache?
                    self.msg_receiver.flush(time.time()-2)
                    flush_count = 0

            except KeyboardInterrupt, e:
                if debug:
                    log_avc.debug("KeyboardInterrupt exception in audit thread")
                thread.interrupt_main()

            except SystemExit, e:
                if debug:
                    log_avc.debug("SystemExit exception in audit thread")
                thread.interrupt_main()

            except Exception, e:
                log_avc.exception("exception %s: %s",e.__class__.__name__, str(e))
                return

