#
# Copyright (C) 2006 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

__all__ = ['log_init',
           'debug',
           'profile',
          ]

import os
import logging
import logging.handlers
import re

from setroubleshoot.config import *

log_init_done = False
debug = False
profile = False

def log_init(program_name):

    # WARNING, for reasons I don't understand log_init is being called multiple
    # times. The only place its called from is the package __init__.py file
    # which I would expect would execute exaclty once per process.
    # This function can only be called once otherwise multiple copies of the
    # logging handlers get created and installed.

    global log_init_done, debug, profile
    if log_init_done:
        return
    log_init_done = True

    program_name = os.path.basename(program_name)
    config_section = '%s_log' % program_name

    pkg_name = cfg.get('general','pkg_name')

    if cfg.has_section(config_section):
        global_level   = map_level(cfg.get(config_section,'level'))
        filename       = cfg.get(config_section,'filename')
        filemode       = cfg.get(config_section,'filemode')
        format         = cfg.get(config_section,'format',raw=True)
        cfg_categories = cfg.get(config_section,'categories')
        console_flag   = cfg.get(config_section,'console').lower() == "true"
        profile        = cfg.get(config_section,'profile').lower() == "true"
    else:
        logging.warning("config section (%s) not found", config_section)
        global_level   = map_level('debug')
        filename       = None
        filemode       = 'w'
        format         = None
        cfg_categories = ''
        console_flag   = False
        profile        = False

    try:
        logging.basicConfig(level=global_level, format=format, filename=filename,
                            filemode=filemode)
    except IOError:
        # Could not open the log file - open the default stderr log
        logging.basicConfig()
        logging.warning("Could not open log file (%s) - using stderr", filename)
        

    # Get the root logger
    root_logger = logging.getLogger()

    # Note: levels NOTSET=0, DEBUG=10, INFO=20, WARNING=30, ERROR=40, CRITICAL=50

    # Set global debug flag
    if global_level < logging.INFO:
        debug = True

    # Iterate through categories defined in the config file.
    # For each category in the config file record the level specified for it
    cfg_loggers = {}
    for name, level in parse_log_categories(cfg_categories):
        cfg_loggers[name] = map_level(level)

    # Iterate over all loggers and create them
    g = globals()
    for name in LOG_CATEGORIES:
        logger_obj_name = "log_%s" % (name)
        logger_obj = logging.getLogger(name)

        # Insert this logger into this modules namespace
        g[logger_obj_name] = logger_obj
        __all__.append(logger_obj_name)

        # If a level was defined for this logger in the config file then set it
        level = cfg_loggers.get(name, None)
        if level:
            logger_obj.setLevel(level)
            # Set global debug flag if any category requests debug messages
            if level < logging.INFO:
                debug = True

    if profile:
        log_stats.setLevel(logging.INFO)

    formatter = logging.Formatter(format)

    # Set up console logging
    if console_flag:
        console = logging.StreamHandler()
        console.setFormatter(formatter)
        root_logger.addHandler(console)

    # Send any log messages at ERROR level or higher to the sys log
    syslog_formatter = logging.Formatter("%s: %s" % (pkg_name, format))

    syslog = logging.handlers.SysLogHandler(address='/dev/log') 
    syslog.setLevel(logging.ERROR)
    syslog.setFormatter(syslog_formatter)
    root_logger.addHandler(syslog)

def map_level(map_str):
    'Parse a string representing a level, map it to a number'
    if map_str is None:
        return None
    try:
        level = int(map_str)
    except ValueError:
        level = logging.getLevelName(map_str.upper())
    return level

def parse_log_categories(str):
    '''Iteratate over a category list in a string,
    return an array of pairs (name,level)'''
    result = []
    category_re = re.compile('^\s*(~\s*)?([^:\s]+)(\s*:\s*([^:\s]+))?\s*$')
    categories = str.split(',')
    for category in categories:
        match = category_re.search(category)
        if match:
            negate = match.group(1)
            name = match.group(2)
            level = match.group(4)
            if negate:
                level = logging.CRITICAL
            result.append((name,level))
    return result


