#
# Copyright (C) 2006 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

import inspect
import libxml2
import re

from types import *
from errno import *
import socket as Socket
import os
import gobject
import threading
import getopt
import sys

from setroubleshoot.config import cfg
from setroubleshoot.errcode import *
from setroubleshoot.log import *
from setroubleshoot.signature import GetElements
from setroubleshoot.util import *

__all__ = [
    'rpc_method',
    'rpc_arg_type',
    'rpc_callback',
    'rpc_signal',
    'interface_registry',
    
    'get_server_address',
    'get_socket_family_and_type',
    'io_input_conditions',

    'RpcManage',
    'RpcChannel',
    'ListeningServer',
    ]

#--------------------------------- Variables ---------------------------------

io_input_conditions = gobject.IO_IN | gobject.IO_HUP | gobject.IO_ERR | gobject.IO_NVAL
content_length_re = re.compile("content-length:(\d+)")
header_end_re = re.compile("\r\n\r\n")
header_field_re = re.compile("([a-zA-Z0-9_-]+):(.*)\r\n")


#------------------------------ Utility Functions -----------------------------

def get_server_address(cfg_section):
    socket_type = cfg.get(cfg_section,'socket_type')
    if socket_type == 'unix':
        path = cfg.get(cfg_section,'path')
        server_address = path
    elif socket_type == 'inet':
        host = cfg.get(cfg_section,'host')
        port = cfg.getint(cfg_section,'port')
        server_address = (host, port)
    else:
        raise ValueError("unknown socket type = %s" % socket_type)
    return server_address

def get_socket_family_and_type(socket_address):
    if type(socket_address) == TupleType:
        address_family = Socket.AF_INET
        socket_type = Socket.SOCK_STREAM
    elif type(socket_address) == StringType:
        address_family = Socket.AF_UNIX
        socket_type = Socket.SOCK_STREAM
    else:
        raise ValueError("unknown socket address (%s)" % socket_address)

    return address_family, socket_type

def io_condition_to_string(io_condition):
    names = []

    if io_condition & gobject.IO_IN:
        names.append('IN')
    if io_condition & gobject.IO_OUT:
        names.append('OUT')
    if io_condition & gobject.IO_PRI:
        names.append('PRI')
    if io_condition & gobject.IO_ERR:
        names.append('ERR')
    if io_condition & gobject.IO_HUP:
        names.append('HUP')
    if io_condition & gobject.IO_NVAL:
        names.append('NVAL')

    return '(%d)[%s]' % (io_condition, ','.join(names))


def rpc_header(body, **kwds):
    hdr = "content-length: %d\r\n" % len(body)
    for key,value in kwds.items():
        hdr += "%s: %s\r\n" % (key, value)
    hdr += "\r\n"
    return hdr

def rpc_message(rpc_id, type, body):
    hdr = rpc_header(body, rpc_id=rpc_id, type=type)
    return hdr+body
    
def convert_rpc_xml_to_args(cmd):
    doc = libxml2.parseDoc(cmd)
    context = doc.xpathNewContext()

    cmd = doc.xpathEval('/cmd')[0]
    interface = cmd.prop('interface')
    method    = cmd.prop('method')

    rpc_def = interface_registry.get_rpc_def(interface, method)

    arg_nodes = context.xpathEval('/cmd/arg')
    args = preextend_list(len(arg_nodes))
    for arg_node in arg_nodes:
        arg_name      = arg_node.prop('name')
        arg_type      = arg_node.prop('type')
        arg_position  = int(arg_node.prop('position'))
        rpc_arg       = rpc_def.positional_args[arg_position]
        if debug:
            log_rpc.debug("arg_name=%s rpc_arg=%s", arg_name, rpc_arg)
        if rpc_arg.obj_type is not None:
            if arg_type == 'xml':
                arg_value = GetElements(arg_node, '*')[0]
                arg_value = rpc_arg.obj_type(arg_value, obj_name=arg_name)
            else:
                arg_value = rpc_arg.obj_type(arg_node.content)
        else:
            arg_value = arg_node.content
        args[arg_position] = arg_value

    doc.freeDoc()
    context.xpathFreeContext()

    return interface, method, args

def convert_rpc_to_xml(rpc_id, rpc_def, *args):
    interface = rpc_def.interface
    method = rpc_def.method

    doc = libxml2.newDoc('1.0')
    root = libxml2.newNode('cmd')
    root.setProp('interface', interface)
    root.setProp('method', method)
    doc.setRootElement(root)

    position = 0
    for rpc_arg in rpc_def.positional_args:
        arg_name = rpc_arg.name
        arg_value = args[position]
        arg_node = libxml2.newNode('arg')
        root.addChild(arg_node)
        arg_node.setProp('name', arg_name)
        arg_node.setProp('position', str(position))
        if isinstance(arg_value, libxml2.xmlNode):
            arg_node.setProp('type', 'xml')
            arg_node.addChild(arg_value)
        elif hasattr(arg_value, 'get_xml_nodes'):
            arg_node.setProp('type', 'xml')
            arg_node.addChild(arg_value.get_xml_nodes(doc, arg_name))
        else:
            arg_node.setProp('type', 'string')
            arg_node.addContent(str(arg_value))
        position += 1
    root.setProp('arg_count',str(position))
    text_doc = doc.serialize(format = 1)
    doc.freeDoc()
    return text_doc

#-----------------------------------------------------------------------------

class RpcArg(object):
    def __init__(self, name=None, obj_type=None):
        self.name = name
        self.obj_type = obj_type

    def __str__(self):
        if self.name is None:
            name = 'name_undefined'
        else:
            name = self.name
        if self.obj_type is None:
            return name
        else:
            return "%s:%s" % (name, self.obj_type)

error_positional_args = [RpcArg('method', str), RpcArg('err_code', int), RpcArg('err_msg', str)]

#-----------------------------------------------------------------------------

class RpcDefinition(object):
    def __init__(self, type, interface, method, rpc_args):
        self.type = type
        self.interface = interface
        self.method = method
        if rpc_args is None:
            self.positional_args = []
        else:
            self.positional_args = rpc_args

    def set_positional_args(self, arg_names):
        if arg_names is not None:
            self.positional_args = preextend_list(len(arg_names), self.positional_args, RpcArg)
            position = 0
            for arg_name in arg_names:
                rpc_arg = self.positional_args[position]
                rpc_arg.name = arg_name
                position += 1
        
    def set_arg_obj_types(self, *obj_types):
        if obj_types is not None:
            self.positional_args = preextend_list(len(obj_types), self.positional_args, RpcArg)
            position = 0
            for obj_type in obj_types:
                rpc_arg = self.positional_args[position]
                rpc_arg.obj_type = obj_type
                position += 1
        
    def get_positional_arg_names(self):
        return [rpc_arg.name for rpc_arg in self.positional_args]

    def __str__(self):
        args = []
        if self.positional_args is not None:
            for rpc_arg in self.positional_args:
                if rpc_arg.obj_type is not None:
                    args.append("%s:%s" % (rpc_arg.name, rpc_arg.obj_type))
                else:
                    args.append(rpc_arg.name)
        return "[%s] %s:%s (%s)" % (self.type, self.interface, self.method, ','.join(args))

#-----------------------------------------------------------------------------

class InterfaceRegistry(object):
    def __init__(self):
        self.interfaces = {}

    def new_interface(self, interface):
        rpc_def = RpcDefinition('error_return', interface, '_error_return', error_positional_args)
        return {'_error_return' : rpc_def}

    def get_interface(self, interface):
        return self.interfaces.setdefault(interface, self.new_interface(interface))

    def set_rpc_def(self, type, interface, method_ptr):
        interface_dict = self.get_interface(interface)
        method = method_ptr.__name__
        positional_args = inspect.getargspec(method_ptr)[0]
        if positional_args[0] == 'self':
            del(positional_args[0])

        rpc_def = self.get_rpc_def(interface, method)
        rpc_def.set_positional_args(positional_args)
        return rpc_def

    def get_rpc_def(self, interface, method):
        interface_dict = self.get_interface(interface)
        if type(method) == MethodType:
            method = method.__name__
        rpc_def = interface_dict.get(method)
        if rpc_def is None:
            rpc_def = RpcDefinition(None, interface, method, None)
            interface_dict[method] = rpc_def
        return rpc_def
        
    def get_error_rpc_def(self, interface):
        interface_dict = self.get_interface(interface)
        return interface_dict['_error_return']

    def dump_interfaces(self):
        interface_names = self.interfaces.keys()
        interface_names.sort()
        for interface_name in interface_names:
            interface = self.interfaces[interface_name]
            print "Interface: %s" % interface_name
            method_names = interface.keys()
            method_names.sort()
            for method_name in method_names:
                method = interface[method_name]
                print "    %s" % str(method)

interface_registry = InterfaceRegistry()

#-------------------------------- Decorators ---------------------------------
    
def rpc_method(interface):
    def decorator(method_ptr):
        rpc_def = interface_registry.set_rpc_def('method', interface, method_ptr)
        method = method_ptr.__name__
        if debug:
            log_rpc.debug("interface=%s method=%s positional_args=%s",
                          rpc_def.interface, method, rpc_def.get_positional_arg_names())
        def rpc_func(self, *args):
            rpc_id = self.new_rpc_id()
            rpc_def = interface_registry.get_rpc_def(interface, method)
            async_rpc = AsyncRpc(rpc_def, rpc_id)
            self.async_rpc_cache[rpc_id] = async_rpc
            self.emit_rpc(rpc_id, 'method', rpc_def, *args)
            return async_rpc
        method_ptr._rpc_definition = True
        return rpc_func
    return decorator

def rpc_arg_type(interface, *arg_types):
    def decorator(method_ptr):
        method = method_ptr.__name__
        rpc_def = interface_registry.get_rpc_def(interface, method)
        rpc_def.set_arg_obj_types(*arg_types)
        if debug:
            log_rpc.debug("interface=%s method=%s arg_types=%s",
                          rpc_def.interface, rpc_def.method, arg_types)
        return method_ptr
    return decorator

def rpc_callback(interface, method):
    def decorator(method_ptr):
        rpc_callback_def = interface_registry.set_rpc_def('method_return', interface, method_ptr)
        if debug:
            log_rpc.debug("interface=%s method=%s positional_args=%s",
                          rpc_callback_def.interface, rpc_callback_def.method, rpc_callback_def.get_positional_arg_names())
        rpc_def = interface_registry.get_rpc_def(interface, method)
        rpc_def.callback = rpc_callback_def.method
        method_ptr._rpc_definition = True
        return method_ptr
    return decorator


def rpc_signal(interface):
    def decorator(method_ptr):
        rpc_def = interface_registry.set_rpc_def('signal', interface, method_ptr)
        method = method_ptr.__name__
        if debug:
            log_rpc.debug("interface=%s method=%s positional_args=%s",
                          rpc_def.interface, method, rpc_def.get_positional_arg_names())
        def rpc_func(self, *args):
            rpc_id = self.new_rpc_id()
            rpc_def = interface_registry.get_rpc_def(interface, method)
            self.emit_rpc(rpc_id, 'signal', rpc_def, *args)
        method_ptr._rpc_definition = True
        return rpc_func
    return decorator

#-----------------------------------------------------------------------------

class GObjectIO(object):
    def __init__(self, on_connection_state_change=None, channel_type=None, channel_name=None):
        if on_connection_state_change is None:
            self.on_connection_state_change = self.default_on_connection_state_change
        else:
            self.on_connection_state_change = on_connection_state_change
        self.connection_state = 'closed'
        self.channel_type = channel_type
        self.channel_name = channel_name
        self.io_watch_id = None

    def set_connection_state(self, connection_state):
        if connection_state not in ['closed', 'open', 'retry', 'hup', 'error', 'invalid', 'timeout']:
            log_rpc.error("set_connection_state: unknown state (%s)" % connection_state)
        if connection_state == self.connection_state:
            return
        self.connection_state = connection_state
	if self.on_connection_state_change:
	    self.on_connection_state_change(self.connection_state)

    def get_connection_state(self):
        return self.connection_state

    def valid_io_condition(self, io_condition):
        if io_condition & (gobject.IO_HUP | gobject.IO_ERR | gobject.IO_NVAL):
            if io_condition & gobject.IO_HUP:
                self.set_connection_state('hup')

            if io_condition & gobject.IO_ERR:
                self.set_connection_state('error')

            if io_condition & gobject.IO_NVAL:
                self.set_connection_state('invalid')

            self.close_connection()
            return False
        else:
            return True
        
    def default_on_connection_state_change(self, connection_state):
        if debug:
            log_rpc.debug("on_connection_state_change: state=%s channel=%s type=%s",
                          connection_state, self.channel_name, self.channel_type)
        if connection_state == 'closed':
            pass
        elif connection_state == 'open':
            pass
        elif connection_state == 'retry':
            pass
        elif connection_state == 'hup':
            pass
        elif connection_state == 'error':
            pass
        elif connection_state == 'invalid':
            pass
        elif connection_state == 'timeout':
            pass
        else:
            log_rpc.error("on_connection_state_change: unknown state (%s) channel=%s type=%s",
                          connection_state, self.channel_name, self.channel_type)


#-----------------------------------------------------------------------------

class ListeningServer(GObjectIO):

    allow_reuse_address = False
    request_queue_size = 5

    def __init__(self, socket_address, client_connection_handler_class, on_connection_state_change=None):
        GObjectIO.__init__(self, on_connection_state_change, 'listening', 'server_listening')
        self.socket_address = socket_address
        self.address_family = None
        self.socket_type = None
        self.socket = None
        self.client_connection_handler_class = client_connection_handler_class

    def new_listening_socket(self, socket_address):
        self.socket_address = socket_address
        self.address_family, self.socket_type = get_socket_family_and_type(socket_address)
        if self.address_family == Socket.AF_UNIX:
            # Unix domain socket, delete the socket if left from before
            if os.path.exists(socket_address):
                os.remove(socket_address)

        self.socket = Socket.socket(self.address_family, self.socket_type)

        if self.allow_reuse_address:
            self.socket.setsockopt(Socket.SOL_SOCKET, Socket.SO_REUSEADDR, 1)
        self.socket.bind(self.socket_address)
        if self.address_family == Socket.AF_UNIX:
            os.chmod(socket_address, 0666)
        self.socket.listen(self.request_queue_size)

        return self.socket


    def open(self):
        self.socket = self.new_listening_socket(self.socket_address)
        self.io_watch_id = gobject.io_add_watch(self.socket, io_input_conditions, self.handle_client_connect)

    def handle_client_connect(self, socket, io_condition):
        if debug:
            #log_rpc.debug("handle_client_connect(): io_condition=%s", io_condition_to_string(io_condition))
            pass

        try:
            if not self.valid_io_condition(io_condition):
                return False

            if io_condition & gobject.IO_IN:
                try:
                    client_socket, client_address = socket.accept()
                    client_handler = self.client_connection_handler_class()
                    client_handler.open(client_socket, client_address)

                except Socket.error, e:
                    log_rpc.error("closing client connection due to socket error: %s", e)
                    self.set_connection_state('error')
                    self.close_connection()

        except Exception, e:
            log_rpc.exception("exception %s: %s",e.__class__.__name__, str(e))
            self.set_connection_state('error')
            self.close_connection()

        return True

#-----------------------------------------------------------------------------

class RequestReceiver:
    def __init__(self, dispatchFunc):
        self.dispatchFunc = dispatchFunc
        self.reset()
    
    def reset(self):
        self.headerLen = -1
        self.bodyLen = 0
        self.header = None
        self.body = ''
        self.feed_buf = ''
        
    def process(self):
        if len(self.feed_buf) == 0:
            # No data, nothing to process
            return
        while True:
            if self.headerLen < 0:
                # Have not seen the end of the header yet, is it there now?
                match = header_end_re.search(self.feed_buf)
                if match:
                    # Yes, header is complete, convert to dict, mark header end
                    self.headerLen = match.end()
                    self.parse_header()
                    continue
                else:
                    # Can't read header till more data arrives
                    break
            if len(self.feed_buf) >= self.headerLen + self.bodyLen:
                # Read entire request, dispatch request, reset for next request
                bodyBegin = self.headerLen
                bodyEnd   = self.headerLen+self.bodyLen
                self.body = self.feed_buf[bodyBegin:bodyEnd]
                self.feed_buf = self.feed_buf[bodyEnd:]
                self.headerLen = -1
                self.bodyLen = -1
                if debug:
                    #log_rpc.debug("dispatch msg: header=%s body=%s", self.header, self.body)
                    pass
                self.dispatchFunc(self.header, self.body)
                continue
            # Have neither a full header nor a full body.
            # Exit till more data is available.
            break

    def feed(self, data):
        self.feed_buf += data
        self.process()

    def parse_header(self):
        self.header = {}
        begin = 0
        while 1:
            match = header_field_re.search(self.feed_buf, begin, self.headerLen+1)
            if match:
                key = match.group(1)
                value = match.group(2).strip()
                self.header[key] = value
                begin = match.end()
            else:
                break
        self.bodyLen = int(self.header['content-length'])

#-----------------------------------------------------------------------------

class RpcManage(object):
    def __init__(self):
	self.async_rpc_cache = {}
        self.rpc_handlers = {}
        self.rpc_id = 0

    def new_rpc_id(self):
        self.rpc_id +=1
        return str(self.rpc_id)
    
    def flush_async_rpc_cache(self):
        self.async_rpc_cache.clear()

    def default_errback(self, method, err_code, err_msg):
        log_rpc.error("[%s] %d %s", method, err_code, err_msg)

    def process_async_return(self, async_rpc):
        if async_rpc is None:
            log_rpc.error("process_async_return(): rpc_id=%s not in async_rpc_cache", rpc_id)
            return
        if async_rpc.return_type == 'method_return':
            for callback in async_rpc.callbacks:
                callback(*async_rpc.return_args)
        elif async_rpc.return_type == 'error_return':
            if len(async_rpc.errbacks) > 0:
                for callback in async_rpc.errbacks:
                    callback(*async_rpc.return_args)
            else:
                self.default_errback(*async_rpc.return_args)

    def connect_rpc_interface(self, interface, handler):
        self.rpc_handlers[interface] = handler
        
#-----------------------------------------------------------------------------

class RpcChannel(GObjectIO, RpcManage):
    socket_buf_size = cfg.getint('socket','buf_size')
    socket_timeout  = cfg.getint('socket','timeout')

    def __init__(self, on_connection_state_change=None, channel_type=None, channel_name=None):
        GObjectIO.__init__(self, on_connection_state_change, channel_type, channel_name)
        RpcManage.__init__(self)
        self.socket_address = None
        self.address_family = None
        self.socket_type = None
        self.socket = None
        self.write_lock = threading.Lock()
        self.receiver = RequestReceiver(self.default_request_handler)

    def __str__(self):
        return "channel: name=%s addr=%s type=%s" % (self.channel_name, self.socket_address, self.channel_type)

    def acquire_write_lock(self):
        self.write_lock.acquire()

    def release_write_lock(self):
        self.write_lock.release()
        
    def set_channel_type(self, channel_type):
        self.channel_type = channel_type
        
    def get_channel_type(self):
        return self.channel_type

    def close_connection(self):
        if self.connection_state == 'closed':
            return
        self.flush_async_rpc_cache()
        if self.socket is not None:
            try:
                self.socket.shutdown(Socket.SHUT_RDWR)
                self.socket.close()
                self.socket = None
            except Socket.error, e:
                self.socket = None

        self.set_connection_state('closed')
        if self.io_watch_id is not None:
            gobject.source_remove(self.io_watch_id)
        self.io_watch_id = None

    def get_method_implementation(self, interface, method):
        handler_obj = self.rpc_handlers.get(interface, None)
        if handler_obj is None:
            return None
        method_ptr = getattr(handler_obj, method, None)
        return method_ptr

    def emit_rpc(self, rpc_id, type, rpc_def, *args):
        rpc_xml = convert_rpc_to_xml(rpc_id, rpc_def, *args)
        rpc_data = rpc_message(rpc_id, type, rpc_xml)
        self.send_data(rpc_data)

    def send_data(self, data):
        if self.connection_state != 'open':
            return
        if debug:
            #log_rpc.debug("send_data() data=%s", data)
            pass
        self.acquire_write_lock()
        try:
            totalSent = 0
            while totalSent < len(data):
                sent = self.socket.send(data[totalSent:])
                if sent == 0:
                    raise RuntimeError, "socket connection broken"
                totalSent = totalSent + sent
        except Socket.timeout, e:
            log_rpc.error("socket timeout (%s)", self.socket_address)
            self.release_write_lock()
            self.set_connection_state('timeout')
            self.close_connection()
            return
        except Socket.error, e:
            log_rpc.error("could send data on socket (%s): %s", self.socket_address, e)
            self.release_write_lock()
            self.set_connection_state('error')
            self.close_connection()
            return
        self.release_write_lock()

    def handle_client_io(self, socket, io_condition):
        if debug:
            #log_rpc.debug("handle_client_io(): io_condition=%s", io_condition_to_string(io_condition))
            pass

        try:
            if not self.valid_io_condition(io_condition):
                return False

            if io_condition & gobject.IO_IN:
                try:
                    data = socket.recv(self.socket_buf_size)
                    if len(data) == 0:
                        self.set_connection_state('hup')
                        self.close_connection()
                        return False
                except Socket.error, e:
                    log_rpc.error("socket error: %s", str(e))
                    self.set_connection_state('error')
                    self.close_connection()
                    return False

                self.receiver.feed(data)
        except Exception, e:
            log_rpc.exception("exception %s: %s",e.__class__.__name__, str(e))
            self.set_connection_state('error')
            self.close_connection()
            return False

        return True

    def handle_return(self, type, rpc_id, body):
        async_rpc = self.async_rpc_cache.pop(rpc_id, None)
        if async_rpc is None:
            log_rpc.error("handle_return(): rpc_id=%s not in async_rpc_cache", rpc_id)
            return
        interface, method, args = convert_rpc_xml_to_args(body)
        async_rpc.return_type = type
        async_rpc.return_args = args
        self.process_async_return(async_rpc)

    def default_request_handler(self, header, body):
	rpc_id    = header.get('rpc_id', 0)
	interface = header.get('interface', None)
	method    = header.get('method', None)
	type      = header.get('type', None)

        if debug:
            #log_rpc.debug("default_request_handler_%s: rpc_id=%s type=%s interface=%s method=%s {%s}",
            #              self.__class__.__name__, rpc_id, type, interface, method, body)
            pass

	if type == 'error_return' or type == 'method_return':
            self.handle_return(type, rpc_id, body)
        elif type == 'method':
            interface, method, args = convert_rpc_xml_to_args(body)
            method_ptr = self.get_method_implementation(interface, method)
            if method_ptr:
                try:
                    return_args = method_ptr(*args)
                    if return_args is not None:
                        return_args = [return_args]
                        rpc_method_def   = interface_registry.get_rpc_def(interface, method)
                        rpc_callback_def = interface_registry.get_rpc_def(interface, rpc_method_def.callback)
                        self.emit_rpc(rpc_id, 'method_return', rpc_callback_def, *return_args)
                except ProgramError, e:
                    rpc_error_def = interface_registry.get_error_rpc_def(interface)
                    self.emit_rpc(rpc_id, 'error_return', rpc_error_def, method, e.errno, e.strerror)
            else:
                err_code = ENOSYS
                err_msg = "method '%s' is not implemented in class '%s'" % (method, self.__class__.__name__)
                rpc_error_def = interface_registry.get_error_rpc_def(interface)
                self.emit_rpc(rpc_id, 'error_return', rpc_error_def, method, err_code, err_msg)
        elif type == 'signal':
            interface, method, args = convert_rpc_xml_to_args(body)
            method_ptr = self.get_method_implementation(interface, method)
            if method_ptr:
                try:
                    method_ptr(*args)
                except ProgramError, e:
                    rpc_error_def = interface_registry.get_error_rpc_def(interface)
            else:
                err_code = ENOSYS
                err_msg = "method '%s' is not implemented in class '%s'" % (method, self.__class__.__name__)
                rpc_error_def = interface_registry.get_error_rpc_def(interface)
                self.emit_rpc(rpc_id, 'error_return', rpc_error_def, method, err_code, err_msg)
        else:
            raise ValueError("unknown type(%s) for %s:%s" % (type, interface, method))

#-----------------------------------------------------------------------------

class AsyncRpc(object):
    def __init__(self, rpc_def, rpc_id):
        self.rpc_def = rpc_def
        self.rpc_id = rpc_id
        self.return_args = None
        self.return_type = None
        self.callbacks = []
        self.errbacks = []
        
    def add_callback(self, callback):
        self.callbacks.append(callback)

    def add_errback(self, errback):
        self.errbacks.append(errback)

#-----------------------------------------------------------------------------


