#
# Copyright (C) 2006 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

__all__ = ['RunFaultServer',
           'get_host_database',
           'send_alert_notification',
          ]

PROGNAME="setroubleshoot"
import gettext
gettext.bindtextdomain(PROGNAME, "/usr/share/locale")
gettext.textdomain(PROGNAME)
try:
    gettext.install(PROGNAME, "/usr/share/locale", 1)
except IOError:
    import __builtin__
    __builtin__.__dict__['_'] = unicode

from setroubleshoot.access_control import *
from setroubleshoot.analyze import *
from setroubleshoot.avc_audit import *
from setroubleshoot.config import cfg
from setroubleshoot.errcode import *
from setroubleshoot.email_alert import *
from setroubleshoot.log import *
from setroubleshoot.rpc import *
from setroubleshoot.rpc_interfaces import *
from setroubleshoot.signature import *
from setroubleshoot.util import *

import errno
import gobject
import os
import pwd
import Queue
import signal
import socket
from stat import *
import sys
import syslog
import threading
from types import *


#------------------------------ Variables -------------------------------

host_database = None
analysis_queue = None

#------------------------------ Utility Functions -----------------------------

def sighandler(signum, frame):
    if debug:
        log_server.debug("received signal=%s", signum)
    import setroubleshoot.config as config
    if signum == signal.SIGHUP:
        log_server.warning("reloading configuration file")
        config.config_init()
        return
    
    sys.exit()

def get_host_database():
    return host_database

# FIXME: this should be part of ClientNotifier
def send_alert_notification(siginfo):
    for client in connection_pool.clients('sealert'):
        client.alert(siginfo)

#-----------------------------------------------------------------------------

class ConnectionPool(object):
    def __init__(self):
        self.client_pool = {}

    def add_client(self, handler):
        if self.client_pool.has_key(handler):
            log_rpc.warning("add_client: client (%s) already in client pool" % handler)
            return
        self.client_pool[handler] = None

    def remove_client(self, handler):
        if not self.client_pool.has_key(handler):
            log_rpc.warning("remove_client: client (%s) not in client pool" % handler)
            return
        del(self.client_pool[handler])

    def clients(self, channel_type=None):
        for client in self.client_pool:
            if channel_type is None:
                yield client
            elif client.channel_type == channel_type:
                yield client

    def close_all(self, channel_type=None):
        for client in self.client_pool(channel_type):
            client.close_connection()

connection_pool = ConnectionPool()

#------------------------------------------------------------------------------

class AlertPluginReportReceiver(PluginReportReceiver):
    def __init__(self, database):
        super(AlertPluginReportReceiver, self).__init__(database)

    def report_problem(self, siginfo):
        siginfo = super(AlertPluginReportReceiver, self).report_problem(siginfo)
        recipients = cfg.get('email','recipients')
        if len(recipients) != 0:
            to_addrs = [addr.strip() for addr in recipients.split(',')]
            for addr in to_addrs:
                action = siginfo.evaluate_filter_for_user(addr)
                if action == "ignore":
                    to_addrs.remove(addr)
                else:
                    self.database.set_filter(self, siginfo.sig, addr, FILTER_ALWAYS)

            if len(to_addrs):
                email_alert(siginfo, to_addrs)
        
        if debug:
            log_alert.debug("sending alert to all clients")
        send_alert_notification(siginfo)

        # FIXME: should this be using our logging objects in log.py?
        syslog.syslog(siginfo.solution.summary + \
                      _(" For complete SELinux messages. run sealert -l %s" % siginfo.local_id ))
        return siginfo

#-----------------------------------------------------------------------------

class ClientConnectionHandler(RpcChannel,
                              SETroubleshootServerInterface,
                              SETroubleshootDatabaseNotifyInterface,
                              SEAlertInterface,
                              ):
    def __init__(self):
        RpcChannel.__init__(self, self.on_connection_state_change, 'sealert')

        self.database = get_host_database()

        self.connect_rpc_interface('SETroubleshootServer', self)
        self.connect_rpc_interface('SETroubleshootDatabase', self.database)

        self.access = ServerAccess()
        self.username = None
        self.uid = None
        self.gid = None

    def on_connection_state_change(self, connection_state):
        if debug:
            log_rpc.debug("on_connection_state_change: state=%s channel=%s type=%s",
                          connection_state, self.channel_name, self.channel_type)
        if connection_state == 'closed':
            connection_pool.remove_client(self)
        elif connection_state == 'open':
            connection_pool.add_client(self)
        elif connection_state == 'retry':
            pass
        elif connection_state == 'hup':
            pass
        elif connection_state == 'error':
            pass
        elif connection_state == 'invalid':
            pass
        elif connection_state == 'timeout':
            pass
        else:
            log_rpc.error("on_connection_state_change: unknown state (%s) channel=%s type=%s",
                          connection_state, self.channel_name, self.channel_type)

    def open(self, socket, socket_address):
        if self.connection_state == 'open':
            return True
        self.socket = socket
        self.socket_address = socket_address
        self.set_connection_state('open')
        self.io_watch_id = gobject.io_add_watch(self.socket, io_input_conditions, self.handle_client_io)


    # ---- SETroubleshootServerInterface Methods ----

    def database_bind(self, database_name):
        host_database = get_host_database()
        if host_database.properties.name == database_name:
            return host_database.properties
        raise ProgramError(ERR_DATABASE_NOT_FOUND, "database (%s) not found" % database_name)
        

    def logon(self, type, username, password):
        if debug:
            log_rpc.debug("logon(%s) type=%s username=%s", self, type, username)
        self.channel_type = type
        self.channel_name = username
        self.user = self.database.get_user(username)
        if self.user is None:
            self.database.add_user(username)

        return None

    # ----  SETroubleshootDatabaseInterface Methods ----

    def delete_signature(self, sig):
        if debug:
            log_rpc.debug("delete_signature: sig=%s", sig)

        siginfo = self.database.delete_signature(sig)
        return None
        
    def evaluate_alert_filter(self, sig, username):
        if debug:
            log_rpc.debug("evaluate_alert_filter: username=%s sig=%s", username, sig)

        action = self.database.evaluate_alert_filter(sig, username)
        return [action]

    def lookup_local_id(self, local_id):
        if debug:
            log_rpc.debug("lookup_local_id: %s", local_id)

        siginfo = self.database.lookup_local_id(local_id)
        return [siginfo]

    def query_alerts(self, criteria):
        if debug:
            log_rpc.debug("query_alerts: criteria=%s", criteria)

        sigs = self.database.query_alerts(criteria)
        return [sigs]

    def set_filter(self, sig, username, filter_type, data = "" ):
        if debug:
            log_rpc.debug("set_filter: username=%s filter_type=%s sig=\n%s",
                          username, filter_type, sig)

        self.database.set_filter(sig, username, filter_type, data)
        return None


    def set_user_data(self, sig, username, item, data):
        if debug:
            log_rpc.debug("set_user_data: username=%s item=%s data=%s sig=\n%s",
                          username, item, data, sig)

        self.database.set_user_data(sig, username, item, data)
        return None

#------------------------------------------------------------------------------

class ClientNotifier(object):
    def __init__(self, connection_pool):
        self.connection_pool = connection_pool

    # ----  SETroubleshootDatabaseNotifyInterface Methods ----

    def signatures_updated(self, type, item):
        for client in self.connection_pool.clients('sealert'):
            client.signatures_updated(type, item)



#------------------------------------------------------------------------------

def RunFaultServer():
    gobject.threads_init()

    # FIXME
    global host_database, analysis_queue

    signal.signal(signal.SIGHUP, sighandler)
    signal.signal(signal.SIGQUIT, sighandler)
    signal.signal(signal.SIGTERM, sighandler)

    #interface_registry.dump_interfaces()

    try:
        # FIXME: should this be using our logging objects in log.py?
        # currently syslog is only used for putting an alert into
        # the syslog with it's id

        pkg_name = cfg.get('general','pkg_name')
        syslog.openlog(pkg_name)

        # Create an object responsible for sending notifications to clients
        client_notifier = ClientNotifier(connection_pool)

        # Create a database local to this host

        host_database = SETroubleshootDatabase(cfg.get('database','filename'), _("Audit Listener"),
                                               notify=client_notifier)

        # Attach the local database to an object which will send alerts
        # specific to this host

        alert_receiver = AlertPluginReportReceiver(host_database)

        # Create a synchronized queue for analysis requests

        analysis_queue = Queue.Queue(0)

        # Create a thread to peform analysis, it takes AVC objects off
        # the the analysis queue and runs the plugins against the
        # AVC. Analysis requests in the queue may arrive from a
        # variety of places; from the audit system, from a log file
        # scan, etc. The disposition of the analysis (e.g. where the
        # results of the analysis are to go) are included in the queued
        # object along with the data to analyze.

        analyze_thread = AnalyzeThread(analysis_queue)
        analyze_thread.setDaemon(True)
        analyze_thread.start()
    
        # Create a thread to receive messages from the audit system.
        # This is a time sensitive operation, the primary job of this
        # thread is to receive the audit message as quickly as
        # possible and return to listening on the audit socket. When
        # it receives a complete audit event it places it in the
        # analysis queue where another thread will process it
        # independently.

        audit_socket_thread = AuditSocketReceiverThread(analysis_queue, alert_receiver)
        audit_socket_thread.setDaemon(True)
        audit_socket_thread.start()

        # Create a server to listen for alert clients and then run.
        listening_server = ListeningServer(get_server_address('local_fault_server'), ClientConnectionHandler)
        listening_server.open()
        main_loop = gobject.MainLoop()
        main_loop.run()

    except KeyboardInterrupt, e:
        if debug:
            log_server.debug("KeyboardInterrupt in RunFaultServer")

    except SystemExit, e:
        if debug:
            log_server.debug("raising SystemExit in RunFaultServer")

    except Exception, e:
        log_rpc.exception("exception %s: %s",e.__class__.__name__, str(e))

if __name__=='__main__':
    RunFaultServer()

