#
# Copyright (C) 2006 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

__all__ = [
    'audit_msg_decode',
    'preextend_list',
    'fmt_obj',
    'format_elapsed_time',
    'format_text',
    'format_msg',
    'remove_linebreaks',
    'remove_html',
    'get_text_default',
    'get_rpm_nvr_from_header',
    'get_rpm_nvr_by_name',
    'get_rpm_nvr_by_file_path',
    'is_hex',
    'split_rpm_nvr',
    'get_plugin_names',
    'load_plugins',
    'get_os_environment',
    'find_program',
    'get_identity',
    'make_database_filepath',

    'TimeStamp',
    ]

import math
import rpm
import textwrap
import sys, os, platform, pwd, socket
import time
import datetime
import glob, imp
import re
import tempfile
from types import *

from rhpl.translate import _
from setroubleshoot.config import cfg
from setroubleshoot.log import *


redhat_release_path='/etc/redhat-release'
text_wrapper = textwrap.TextWrapper()
fix_newline_re = re.compile("\s*\n+\s*")
hex_re = re.compile('^[A-Fa-f0-9]+$')
href_re = re.compile('<a\s*href="([^"]+)"[^<]*</a>')


def format_elapsed_time(elapsed_time):
    if elapsed_time is None:
        return None

    fraction, whole = math.modf(elapsed_time)
    whole = int(whole)

    days = whole/86400
    whole = whole - days*86400

    hours = whole/3600
    whole = whole - hours*3600
    
    minutes = whole/60
    seconds = whole - minutes*60
    
    if days:
        return "%dd:%dh:%dm:%.3fs" % (days,hours,minutes,seconds+fraction)
    if hours:
        return "%dh:%dm:%.3fs" % (hours,minutes,seconds+fraction)
    if minutes:
        return "%dm:%.3fs" % (minutes,seconds+fraction)
    return "%.3fs" % (seconds+fraction)

def is_hex(str):
    if hex_re.match(str):
        return True
    else:
        return False

def audit_msg_decode(msg):
    if msg is None:
        return None
    msg = msg.strip('"')
    if is_hex(msg):
        msg = msg.decode('hex')
    return msg



def preextend_list(requested_length, _list=None, default=None):
    if _list is None:
        _list = []
    cur_length = len(_list)
    delta = requested_length-cur_length
    if delta > 0:
        if type(default) is TypeType:
            _list.extend([default() for x in range(delta)])
        else:
            _list.extend([default] * delta)
    return _list

def fmt_obj(obj):
    if type(obj) is StringType:
        return obj
    elif type(obj) is ListType or type(obj) is TupleType:
        return "["+" ".join(["%s" % fmt_obj(x) for x in obj])+"]"
    elif type(obj) is DictType:
        keys = obj.keys()
        keys.sort()
        return "{"+" ".join(["%s=%s" % (fmt_obj(key), fmt_obj(obj[key])) for key in keys])+"}"
    else:
        return str(obj)

def format_text(s):
    return text_wrapper.fill(s)

def format_msg(title, msg, indent=4):
    if msg is None:
        msg = ''
    msg = msg.strip()
    indentString = " " * indent
    text_wrapper.initial_indent    = indentString
    text_wrapper.subsequent_indent = indentString
    text_wrapper.width = 80
    return title + "\n" + text_wrapper.fill(msg) + "\n\n"

def remove_linebreaks(str):
    new_str = fix_newline_re.sub(" ", str).strip()
    if new_str is None:
        return ""
    else:
        return new_str


# WARNING: this is not a complete or robust implementation!
def remove_html(msg):
    msg = remove_linebreaks(msg)
    msg = href_re.sub('\\1', msg)
    return msg

def get_text_default(val):
    if val is None:
        val = _('Unknown')
    return val

def get_rpm_nvr_from_header(hdr):
    'Given an RPM header return the package NVR as a string'
    name    = hdr['name']
    version = hdr['version']
    release = hdr['release']

    return "%s-%s-%s" % (name, version, release)

def get_rpm_nvr_by_name(name):
    if name is None:
        return None

    nvr = None
    try:
        ts = rpm.ts()
        mi = ts.dbMatch(rpm.RPMTAG_NAME, name)
        for h in mi:
            nvr = get_rpm_nvr_from_header(h)
            break
    except:
        log_plugin.exception("failed to retrieve rpm info for %s", name)
    return nvr

def get_rpm_nvr_by_file_path(path):
    if path is None:
        return None

    nvr = None
    try:
        ts = rpm.ts()
        mi = ts.dbMatch(rpm.RPMTAG_BASENAMES, path)
        for h in mi:
            nvr = get_rpm_nvr_from_header(h)
            break
    except:
        log_plugin.exception("failed to retrieve rpm info for %s", path)
    return nvr

def split_rpm_nvr(nvr):
    components = nvr.split('-')
    release = components[-1]
    version = components[-2]
    name    = '-'.join(components[:-2])
    return (name, version, release)


#------------------------------------------------------------------------------

STDOFFSET = datetime.timedelta(seconds = -time.timezone)
if time.daylight:
    DSTOFFSET = datetime.timedelta(seconds = -time.altzone)
else:
    DSTOFFSET = STDOFFSET

DSTDIFF = DSTOFFSET - STDOFFSET
ZERO = datetime.timedelta(0)
HOUR = datetime.timedelta(hours=1)


# A class capturing the platform's idea of local time.
class LocalTimezone(datetime.tzinfo):

    def utcoffset(self, dt):
        if self._isdst(dt):
            return DSTOFFSET
        else:
            return STDOFFSET

    def dst(self, dt):
        if self._isdst(dt):
            return DSTDIFF
        else:
            return ZERO

    def tzname(self, dt):
        return time.tzname[self._isdst(dt)]

    def _isdst(self, dt):
        tt = (dt.year, dt.month, dt.day,
              dt.hour, dt.minute, dt.second,
              dt.weekday(), 0, -1)
        stamp = time.mktime(tt)
        tt = time.localtime(stamp)
        return tt.tm_isdst > 0

class UTC(datetime.tzinfo):
    """UTC"""

    def utcoffset(self, dt):
        return datetime.timedelta(0)

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return datetime.timedelta(0)

class TimeStamp:
    # class variables
    utc_tz = UTC()
    local_tz = LocalTimezone()

    iso8601_fmt = '%Y-%m-%dT%H:%M:%SZ'
    ctime_fmt = '%a %b %d %H:%M:%S %Y'
    time_only_fmt = '%l:%M %p'
    weekday_fmt = '%a %l:%M %p'
    this_year_fmt = '%b %l:%M %p'
    older_than_a_year_fmt = '%b %d %Y'
    future_fmt = '%a %l:%M %p %Y'

    def __init__(self, t=None):
        if t is None:
            self.dt = self.now(local=False)
        elif type(t) is StringType:
            self.parse(t)
        elif type(t) is FloatType:
            self.dt = datetime.datetime.fromtimestamp(t, self.utc_tz)
        elif isinstance(t, datetime.datetime):
            self.dt = t
    
    def now(self, local=False):
        if local:
            return datetime.datetime.now(self.local_tz)
        else:
            return datetime.datetime.now(self.utc_tz)
    
    def local(self):
        return self.dt.astimezone(self.local_tz)

    def __str__(self):
        return self.format(self.iso8601_fmt, local=False)

    def parse(self, str):
        (year, month, day, hour, minute, second, weekday, yearday, dst) = \
               time.strptime(str, self.iso8601_fmt)
        self.dt = datetime.datetime(year, month, day, hour, minute, second,
                                    0, self.utc_tz)
        return self.dt

    def add(self,days=0, hours=0, minutes=0, seconds=0):
        self.dt += datetime.timedelta(days=days, hours=hours,
                                      minutes=minutes, seconds=seconds)
        
    def cmp(self, ts):
        if self.dt < ts.dt:
            return -1
        elif self.dt > ts.dt:
            return 1
        else:
            return 0

    def in_future(self):
        now = self.now()
        if now < self.dt:
            return True
        else:
            return False

    def in_past(self):
        now = self.now()
        if now >= self.dt:
            return True
        else:
            return False

    def format(self, fmt=None, local=True):
        if fmt is None:
            fmt = self.ctime_fmt
        if local:
            return self.local().strftime(fmt)
        else:
            return self.dt.strftime(fmt)
        
    def friendly_format(self, local=True):
        if local:
            dt = self.local()
            now = self.now(local=True)
        else:
            dt = self.dt
            now = self.now(local=False)

        # Note: because of the way timedelta's are normalized for negative values
        # we make positive deltas be in the past, negative values in the future.
        delta = now - dt
        if delta.days < 0:
            mystr = dt.strftime(self.future_fmt)
        elif delta.days == 0:
            mystr = dt.strftime(self.time_only_fmt)
        elif delta.days == 1:
            mystr = '%s %s' % (_('Yesterday'), dt.strftime(self.time_only_fmt))
        elif delta.days < 7:
            mystr = dt.strftime(self.weekday_fmt)
        elif delta.days < 365:
            mystr = dt.strftime(self.this_year_fmt)
        else:
            mystr = dt.strftime(self.older_than_a_year_fmt)
        return mystr

#------------------------------------------------------------------------------

def get_plugin_names(filter_glob=None):
    if filter_glob is None:
        filter_glob = '*'
    else:
        filter_glob = re.sub('.py$','',filter_glob)

    plugin_dir = cfg.get('plugins','plugin_dir')
    plugin_names = []
    for p in glob.glob(os.path.join(plugin_dir, filter_glob+".py")):
        p = os.path.basename(p)
        if p in ['__init__.py']:
            continue
        plugin_name = os.path.splitext(os.path.basename(p))[0]
        plugin_names.append(plugin_name)
    return plugin_names

def sort_plugins(x,y):
    return x.get_priority()-y.get_priority()
    
def load_plugins(filter_glob=None):
    if debug:
        log_plugin.debug("load_plugins()")
    plugin_dir = cfg.get('plugins','plugin_dir')
    plugin_base = os.path.basename(plugin_dir)

    plugin_names = get_plugin_names(filter_glob)
    if debug:
        log_plugin.debug("load_plugins() names=%s", plugin_names)
    plugins = []
    for plugin_name in plugin_names:
        module_name = "%s.%s" % (plugin_base, plugin_name)
        mod = sys.modules.get(module_name)
        if mod is not None:
            if debug:
                log_plugin.debug("load_plugins() %s previously imported", module_name)
            plugins.append(mod.plugin())
            continue
        log_plugin.info("importing %s as %s",
                         os.path.join(plugin_dir, plugin_name), module_name)
        try:
            mod = imp.load_module(module_name, *imp.find_module(plugin_name, [plugin_dir]))
            plugins.append(mod.plugin())
        except Exception:
            log_plugin.exception("failed to load %s plugin", plugin_name)
    
    plugins.sort(sort_plugins)
    return plugins
                
def get_os_environment():
    try:
        myplatform = open(redhat_release_path).readlines()[0].strip()
    except IOError:
        # dist returns (distname, version, id)
        myplatform = ' '.join(platform.dist())
        
    # uname returns (sysname, nodename, release, version, machine)
    uname = os.uname()
    kernel_release = uname[2]
    cpu            = uname[4]

    os_desc = "%s %s" % (kernel_release, cpu)
    return (myplatform, os_desc)
    
def get_identity():
    uid = os.getuid()
    username = pwd.getpwuid(uid)[0]

    hostname = socket.gethostname()

    identity = "%s@%s" % (username, hostname)
    return identity

def find_program(prog):
    if os.path.isabs(prog):
        return prog
    basename = os.path.basename(prog)
    search_path = cfg.get('fix_command','prog_search_path').split(':')
    for d in search_path:
        path = os.path.join(d, basename)
        if os.path.exists(path):
            return path
    return None

def make_database_filepath(name, temp=False):
    if temp:
        database_dir = tempfile.gettempdir()
    else:
        database_dir = cfg.get('database','database_dir')
    # strip off extension if one was provided
    name = re.sub('\\.xml$', '', name)
    filename = name + '_database.xml'
    filepath = os.path.join(database_dir, filename)
    return filepath

